const path = require('path');
const fs = require('fs');
const childProcess = require('child_process');
const tmp = require('tmp');
const yarnlock = require('@yarnpkg/lockfile');

const autoprefixer = require('autoprefixer');
const CopyWebpackPlugin = require('copy-webpack-plugin');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const HtmlWebpackPlugin = require('html-webpack-plugin');
const { CleanWebpackPlugin } = require('clean-webpack-plugin');
const webpack = require('webpack');
const ForkTsCheckerWebpackPlugin = require('fork-ts-checker-webpack-plugin');
const BundleAnalyzerPlugin = require('webpack-bundle-analyzer').BundleAnalyzerPlugin;
const { DuplicatesPlugin } = require('inspectpack/plugin');
const ReactCMFWebpackPlugin = require('@talend/react-cmf-webpack-plugin');

const AppLoader = require('@talend/react-components/lib/AppLoader/constant').default;
const { getBabelConfig } = require('@talend/scripts-config-babel/babel-resolver');

const cdn = require('@talend/scripts-config-cdn');
const LICENSE_BANNER = require('./licence');
const tsConfigExists = require('./utils/ts-config-exists');
const inject = require('./inject');
const icons = require('./icons');

const INITIATOR_URL = process.env.INITIATOR_URL || '@@INITIATOR_URL@@';

const DEFAULT_INDEX_TEMPLATE_PATH = 'src/app/index.html';
const ICON_DIST = icons.getIconsDistPath();
const getFileNameForExtension = (extension, prefix) =>
	`${prefix ? prefix : ''}[name]-[hash].${extension}`;

// Check if Typescript is setup
const useTypescript = tsConfigExists();

const babelConfig = getBabelConfig();

function getSassData(userSassData) {
	let sassData = ["@import '~@talend/bootstrap-theme/src/theme/guidelines';"];

	if (userSassData && userSassData.data) {
		sassData = Object.keys(userSassData.data)
			.map(key => `${key}: ${userSassData.data[key]};`)
			.concat(sassData);
	}

	return sassData.join('\n');
}

function getCommonStyleLoaders(enableModules, mode) {
	let cssOptions = {
		sourceMap: true,
	};
	if (enableModules) {
		cssOptions = {
			sourceMap: true,
			modules: {
				localIdentName: '[name]__[local]___[hash:base64:5]',
			},
			importLoaders: 1,
		};
	}
	return [
		mode === 'development'
			? { loader: 'style-loader' }
			: { loader: MiniCssExtractPlugin.loader, options: { esModule: false } },
		{ loader: 'css-loader', options: cssOptions },
		{
			loader: 'postcss-loader',
			options: {
				plugins: () => [autoprefixer()],
				sourceMap: true,
			},
		},
		{ loader: 'resolve-url-loader', options: { sourceMap: true } },
	];
}

function getSassLoaders(enableModules, sassData, mode) {
	return getCommonStyleLoaders(enableModules, mode).concat({
		loader: 'sass-loader',
		options: { sourceMap: true, prependData: sassData },
	});
}

const BLOCKED_LIST = ['@talend/backend-mock', '@talend/html-webpack-plugin', '@talend/copylib'];
const BLOCKED_SUBSTRING = ['scripts', 'webpack'];

function isBlockedLib(name) {
	if (BLOCKED_LIST.indexOf(name) !== -1) {
		return true;
	}
	let blocked = false;
	BLOCKED_SUBSTRING.forEach(pattern => {
		if (!blocked) {
			blocked = name.includes(pattern);
		}
	});
	return blocked;
}

function getVersions() {
	const talendUi = [
		'@talend/bootstrap-theme',
		'@talend/icons',
		'@talend/react-cmf',
		'@talend/react-cmf-cqrs',
		'@talend/react-components',
		'@talend/react-containers',
		'@talend/react-forms',
		'@talend/react-sagas',
	];
	const talendLibraries = [];
	const packagelockPath = path.join(process.cwd(), 'package-lock.json');
	const yarnlockPath = path.join(process.cwd(), 'yarn.lock');
	const packageJson = require(path.join(process.cwd(), 'package.json'));
	let talendUiVersion;

	if (fs.existsSync(yarnlockPath)) {
		const data = fs.readFileSync(yarnlockPath, 'utf-8');
		const lock = yarnlock.parse(data);
		Object.keys(lock.object)
			.filter(k => k.startsWith('@talend'))
			.reduce((acc, key) => {
				// @talend/react-components@5.1.2
				const name = `@talend/${key.split('/')[1].split('@')[0]}`;
				const info = lock.object[key];
				if (talendUi.includes(name)) {
					talendUiVersion = info.version;
				} else if (!isBlockedLib(name)) {
					acc.push({
						name,
						version: info.version,
					});
				}
				return acc;
			}, talendLibraries);
	} else if (fs.existsSync(packagelockPath)) {
		const packageLock = require(packagelockPath);
		Object.keys(packageLock.dependencies)
			.filter(k => k.startsWith('@talend'))
			.reduce((acc, key) => {
				const name = `@talend/${key.split('/')[1].split('@')[0]}`;
				const info = packageLock.dependencies[key];
				if (talendUi.includes(name)) {
					talendUiVersion = info.version;
				} else if (!isBlockedLib(name)) {
					acc.push({
						name: key,
						version: info.version,
					});
				}
				return acc;
			}, talendLibraries);
	}

	if (talendUiVersion) {
		talendLibraries.push({
			name: '@talend/ui',
			version: talendUiVersion,
		});
	}

	let revision = process.env.GIT_COMMIT;
	if (!revision) {
		try {
			revision = childProcess.execSync('git rev-parse HEAD').toString().trim();
		} catch (e) {
			console.info('Failed to get git revision');
		}
	}

	return {
		version: packageJson.version,
		talendLibraries,
		revision,
	};
}
const VERSIONS = getVersions();
// meta for html webpack plugin
const meta = VERSIONS.talendLibraries.reduce(
	(acc, lib) => {
		acc[lib.name] = lib.version;
		return acc;
	},
	{
		'app-version': VERSIONS.version,
		'app-revision': VERSIONS.revision,
	},
);

cdn.init();

function getCopyConfig(env, userCopyConfig = []) {
	const config = [...userCopyConfig];
	const assetsOverridden = config.some(nextAsset =>
		typeof nextAsset === 'object' ? nextAsset.from === 'src/assets' : nextAsset === 'src/assets',
	);
	if (!assetsOverridden && fs.existsSync(path.join(process.cwd(), 'src/assets'))) {
		config.push({ from: 'src/assets' });
	}
	config.push({ from: path.join(ICON_DIST, 'svg-bundle') });
	cdn.getCopyConfig().forEach(c => config.push(c));
	return config;
}

function getIndexTemplate(env, mode, indexTemplatePath) {
	/*
	 * The following header is used to inject.
	 * Some UMD are not built correctly for browser.
	 * For example react-is index.js includes a test on process.env.NODE_ENV to require the min version or not.
	 * Let's bypass this issue by setting a process.env.NODE_ENV
	 */
	const header = `<% var meta = htmlWebpackPlugin.options.meta;
		var metaKeys = Object.keys(meta);
		for(var i = 0; i < metaKeys.length; ++i) {%>
		<meta name="<%=metaKeys[i] %>" content="<%=meta[metaKeys[i]] %>" />
		<% } %>
		<link rel="icon" type="image/svg+xml" href="<%= htmlWebpackPlugin.options.appfavicon %>">
		<style><%= htmlWebpackPlugin.options.appLoaderStyle %></style>
		<script type="text/javascript">
			var process = { browser: true, env: { NODE_ENV: '${mode}' } };
			var TALEND_CDN_VERSIONS = {
				TUI: '@@TALEND_UI_VERSION@@', ${process.env.CUSTOM_VERSIONS || ''}
			};
			window.TALEND_INITIATOR_URL = '${INITIATOR_URL}';
			window.jsFiles = [<%= htmlWebpackPlugin.files.js.map(href => '"'+href+'"').join(',') %>];
			window.cssFiles = [<%= htmlWebpackPlugin.files.css.map(href => '"'+href+'"').join(',') %>];
			window.Talend = { build: <%= JSON.stringify(htmlWebpackPlugin.files.jsMetadata)%>, cssBuild:  <%= JSON.stringify(htmlWebpackPlugin.files.cssMetadata)%> };
			${inject.getMinified()}
		</script>
	</head>`;

	const indexTemplate = fs.readFileSync(indexTemplatePath, 'utf8').replace('</head>', header);

	const tmpobj = tmp.fileSync();
	fs.writeFileSync(tmpobj.name, indexTemplate);
	return tmpobj.name;
}

module.exports = ({ getUserConfig, mode }) => {
	return (env = {}) => {
		const cssModulesEnabled = getUserConfig(['css', 'modules'], true);
		const cssPrefix = getUserConfig(['css', 'prefix']);
		const jsPrefix = getUserConfig(['js', 'prefix']);
		const userHtmlConfig = getUserConfig('html', {});
		const userSassData = getUserConfig('sass', {});
		const userCopyConfig = getUserConfig('copy', []);
		const cmf = getUserConfig('cmf');
		const dcwpConfig = getUserConfig('dynamic-cdn-webpack-plugin');
		const { theme } = userSassData;

		const appLoaderIcon = icons.getAppLoaderIconUrl(theme);

		const sassData = getSassData(userSassData);
		const indexTemplatePath = path.join(
			process.cwd(),
			userHtmlConfig.template || DEFAULT_INDEX_TEMPLATE_PATH,
		);
		const indexTemplate = getIndexTemplate(env, mode, indexTemplatePath);

		const isEnvDevelopment = mode === 'development';
		const isEnvProduction = mode === 'production';
		const favicon = icons.getFavicon(theme);

		meta['app-id'] = userHtmlConfig.appId || theme;

		return {
			entry: {
				polyfills: [
					'regenerator-runtime',
					'core-js-bundle',
					isEnvDevelopment && path.join(__dirname, 'wdyr.js'),
				].filter(Boolean),
				theme: [
					'@talend/bootstrap-theme',
					theme && `@talend/bootstrap-theme/src/theme/variations/_${theme}.scss`,
				].filter(Boolean),
				app: [`${process.cwd()}/src/app/index`],
			},
			output: {
				filename: getFileNameForExtension('js', jsPrefix),
				chunkFilename: getFileNameForExtension('js', jsPrefix),
				publicPath: '/',
				globalObject: 'this',
			},
			devtool: 'cheap-module-source-map',
			resolve: {
				extensions: ['.js', useTypescript && '.ts', useTypescript && '.tsx'].filter(Boolean),
			},
			module: {
				rules: [
					isEnvDevelopment && {
						test: /\.js$/,
						include: /@talend/,
						use: ['source-map-loader'],
						enforce: 'pre',
					},
					{
						test: useTypescript ? /\.(js|ts|tsx)$/ : /\.js$/,
						exclude: /node_modules/,
						use: [
							!process.env.NO_CACHE_LOADER && { loader: 'cache-loader' },
							{
								loader: 'babel-loader',
								options: babelConfig,
							},
							useTypescript && {
								loader: 'ts-loader',
								options: {
									// disable type checker - it is done in another process by ForkTsCheckerWebpackPlugin
									transpileOnly: true,
								},
							},
						].filter(Boolean),
					},
					{
						test: /\.html$/,
						use: [
							!process.env.NO_CACHE_LOADER && { loader: 'cache-loader' },
							{
								loader: 'html-loader',
								options: {
									minimize: true,
									removeComments: true,
									collapseWhitespace: true,
								},
							},
						].filter(Boolean),
						exclude: indexTemplatePath,
					},
					{
						test: /\.css$/,
						use: getCommonStyleLoaders(false, mode),
						exclude: /@talend/,
					},
					{
						test: /\.scss$/,
						use: getSassLoaders(false, sassData, mode),
						include: /bootstrap-theme/,
					},
					{
						test: /\.scss$/,
						use: getSassLoaders(true, sassData, mode),
						include: /@talend/,
						exclude: /bootstrap-theme/,
					},
					{
						test: /\.scss$/,
						use: getSassLoaders(cssModulesEnabled, sassData, mode),
						exclude: /@talend/,
					},
					{
						test: /\.woff(2)?(\?v=\d+\.\d+\.\d+)?$/,
						loader: 'url-loader',
						options: {
							name: './fonts/[name].[ext]',
							limit: 10000,
							mimetype: 'application/font-woff',
						},
					},
					{
						test: /\.svg$/,
						loader: 'url-loader',
						options: {
							name: 'assets/svg/[name].[ext]',
							limit: 10000,
							mimetype: 'image/svg+xml',
						},
					},
					{
						test: /\.(png|jpg|jpeg|gif)$/,
						loader: 'url-loader',
						options: {
							name: 'assets/img/[name].[ext]',
							limit: 10000,
							mimetype: 'image/png',
						},
					},
				].filter(Boolean),
			},
			plugins: [
				isEnvDevelopment && new DuplicatesPlugin(),
				new CleanWebpackPlugin({
					verbose: true,
					cleanOnceBeforeBuildPatterns: [path.join(process.cwd(), 'dist')],
				}),
				new webpack.DefinePlugin({
					BUILD_TIMESTAMP: Date.now(),
					TALEND_APP_INFO: JSON.stringify(getVersions()),
					'process.env.ICON_BUNDLE': JSON.stringify(process.env.ICON_BUNDLE),
					'process.env.FORM_MOZ': JSON.stringify(process.env.FORM_MOZ),
					'process.env.DISABLE_JS_ERROR_NOTIFICATION': JSON.stringify(
						process.env.DISABLE_JS_ERROR_NOTIFICATION,
					),
				}),
				isEnvDevelopment &&
					new BundleAnalyzerPlugin({
						analyzerMode: 'static',
						openAnalyzer: !!env.analyze,
						logLevel: 'error',
					}),
				isEnvDevelopment && new webpack.HotModuleReplacementPlugin(),
				isEnvProduction &&
					new MiniCssExtractPlugin({
						filename: getFileNameForExtension('css', cssPrefix),
						chunkFilename: getFileNameForExtension('css', cssPrefix),
					}),
				new HtmlWebpackPlugin({
					filename: './index.html',
					appfavicon: favicon,
					template: indexTemplate,
					appLoader: AppLoader.APP_LOADER,
					appLoaderStyle: AppLoader.getLoaderStyle(appLoaderIcon),
					inject: false,
					meta,
					...userHtmlConfig,
				}),
				cdn.getWebpackPlugin(env, dcwpConfig),
				new CopyWebpackPlugin({ patterns: getCopyConfig(env, userCopyConfig) }),
				new webpack.BannerPlugin({ banner: LICENSE_BANNER }),
				cmf && new ReactCMFWebpackPlugin({ watch: isEnvDevelopment }),
				useTypescript && new ForkTsCheckerWebpackPlugin(),
			].filter(Boolean),
			optimization: {
				// Automatically split vendor and commons
				// https://twitter.com/wSokra/status/969633336732905474
				// https://medium.com/webpack/webpack-4-code-splitting-chunk-graph-and-the-splitchunks-optimization-be739a861366
				splitChunks: {
					chunks(chunk) {
						// exclude `polyfills` chunk as we want to only load it for old browsers
						return chunk.name !== 'polyfills';
					},
				},
				// Keep the runtime chunk separated to enable long term caching
				// https://twitter.com/wSokra/status/969679223278505985
				runtimeChunk: true,
			},
			watchOptions: {
				aggregateTimeout: 300,
				poll: 1000,
			},
			devServer: {
				port: 3000,
				stats: 'errors-only',
				historyApiFallback: true,
				contentBase: path.join(process.cwd(), 'dist'),
				compress: true,
				hot: true,
			},
		};
	};
};
